"""
Module implements base classes for api clients.
"""
import asyncio
from numbers import Real
from typing import Awaitable, Dict, Optional, Union

import aiohttp
import ujson

from .logger import logger
from .luna_response import LunaResponse
from .requests import DNSResolver, RequestPayload, makeRequest

MISSING = object()


class BaseAPI:
    """
    Base class for request to luna services.

    Attributes:
        api (int): api version of  luna, default 1
        asyncRequest (bool): default mode for request async or blocking
        lunaRequestId (str): Luna-Request-Id.
        totalTimeout (Real): total request timeout
        echo (bool): log request and response or not
    """

    def __init__(
        self,
        api: int = 1,
        asyncRequest: bool = False,
        lunaRequestId: Optional[str] = None,
        totalTimeout: Real = 60,
        echo: bool = False,
    ):
        self.api = api
        self.asyncRequest = asyncRequest
        self.lunaRequestId = lunaRequestId

        self.totalTimeout = totalTimeout
        self.echo = echo

    def updateSettings(self, **kwargs):
        """
        Update settings.

        Keyword Args:
            see Attributes
        """
        for setting in self.__dict__:
            if setting in kwargs and kwargs[setting] is not None:
                self.__dict__[setting] = kwargs[setting]

    def getRequestId(self, lunaRequestId: Optional[str] = None) -> Optional[str]:
        """
        Get final request id.

        Args:
            lunaRequestId: Luna-Request-Id as function argument. Pattern:
                ^[0-9]{10},[a-f0-9]{8}-[a-f0-9]{4}-4[a-f0-9]{3}-[89ab][a-f0-9]{3}-[a-f0-9]{12}$
        Returns:
            None if both lunaRequestId and self.lunaRequestId are None or
                lunaRequestId or self.lunaRequestId
        """
        return lunaRequestId if lunaRequestId is not None else self.lunaRequestId

    def getRequestIdAsDict(self, lunaRequestId: Optional[str] = None) -> Dict[str, str]:
        """
        Get final request id.

        Args:
            lunaRequestId: Luna-Request-Id as function argument. Pattern:
                ^[0-9]{10},[a-f0-9]{8}-[a-f0-9]{4}-4[a-f0-9]{3}-[89ab][a-f0-9]{3}-[a-f0-9]{12}$
        Returns:
            empty dict if both lunaRequestId and self.lunaRequestId are None or
                return dict with lunaRequestId or self.lunaRequestId
        """
        requestId = self.getRequestId(lunaRequestId)
        if requestId is not None:
            return {"Luna-Request-Id": requestId}
        return {}

    def getAsyncMode(self, asyncRequest: Optional[bool] = None) -> bool:
        """
        Get final async mode.

        Args:
            asyncRequest: async mode as function argument.

        Returns:
            If async is None, return self.asyncRequest else async
        """
        return asyncRequest if asyncRequest is not None else self.asyncRequest

    def getTotalTimeout(self, totalTimeout: Optional[Real] = None) -> Real:
        """
        Get final totalTimeout.

        Args:
            totalTimeout: request's processing timeout in seconds.

        Returns:
            request timeout seconds.
        """
        return totalTimeout if totalTimeout is not None else self.totalTimeout


class BaseHTTPAPI(BaseAPI):
    """
    Base class for request to luna services.

    Attributes:
        origin (str): protocol, host and port; for example, "http://127.0.0.1:8000"
        api (int): api version of  luna, default 1
        asyncRequest (bool): default mode for request async or blocking
        lunaRequestId (str): Luna-Request-Id.
        totalTimeout (Real): request processing timeout in seconds
        connectTimeout (Real): request socket connect timeout or obtaining connection from session connection pool
        sockReadTimeout (Real): request socket read timeout
        sockConnectTimeout (Real): request socket connect timeout

        _session: session to work in; is created automatically if 'asyncRequest'
        echo: log a request and response or not
        initSession: whether or not to initiate a session now. This is required for background clients created once

    References:
        https://docs.aiohttp.org/en/stable/client_quickstart.html#timeouts
    """

    # multipart boundary; real property is defined above (BaseLunaHttpClient)
    multipartBoundary: str = "LUNACOREB0UnDaRyStRiNg"

    def __init__(
        self,
        origin: str,
        api: int = 1,
        asyncRequest: bool = False,
        lunaRequestId: Optional[str] = None,
        totalTimeout: Real = 60,
        connectTimeout: Optional[Real] = None,
        sockReadTimeout: Optional[Real] = None,
        sockConnectTimeout: Optional[Real] = None,
        *,
        session: Optional[aiohttp.ClientSession] = None,
        initSession: bool = False,
        echo: bool = False,
    ):
        super().__init__(api, asyncRequest, lunaRequestId, totalTimeout, echo=echo)
        self.origin = origin

        self.connectTimeout = connectTimeout
        self.sockReadTimeout = sockReadTimeout
        self.sockConnectTimeout = sockConnectTimeout
        self._session = session
        if not session and initSession:
            self._session = self.session

    @property
    def session(self) -> Union[None, aiohttp.ClientSession]:
        """
        Aiohttp session getter. Same for one BaseHTTPAPI instance.

        Returns:
            initialised aiohttp session or None
        """
        if self._session is None:
            self._session = aiohttp.ClientSession(
                headers=self.getRequestIdAsDict(),
                json_serialize=ujson.dumps,
                timeout=aiohttp.ClientTimeout(
                    total=self.totalTimeout,
                    connect=self.connectTimeout,
                    sock_read=self.sockReadTimeout,
                    sock_connect=self.sockConnectTimeout,
                ),
                connector=aiohttp.TCPConnector(resolver=DNSResolver()),
                cookie_jar=aiohttp.DummyCookieJar(),
            )
        return self._session

    @session.setter
    def session(self, newSession: aiohttp.ClientSession) -> None:
        """
        Session setter.

        Args:
            newSession: session to set
        """
        self._session = newSession

    async def close(self) -> None:
        """Close current client."""
        if self._session is not None:
            await self._session.close()
            self._session = None

    def _connectionChecker(self, reply: LunaResponse) -> bool:
        """
        Checking connection

        Args:
            reply: LunaResponse
        Returns:
            status in bool
        """
        if reply.success and int(reply.json["Version"]["api"]) == self.api:
            return True
        if not reply.success:
            logger.warning(f"Connection check is failed, " f"status code: {reply.statusCode}, error: {reply.json}")
        else:
            logger.warning(
                f"Connection check is failed, service "
                f"version {reply.json['Version']['api']} "
                f"does not match with version {self.api} from settings"
            )
        return False

    async def _asyncTestConnection(
        self,
        lunaRequestId: Optional[str] = None,
        raiseError: Optional[bool] = False,
        totalTimeout: Optional[Real] = None,
        connectTimeout: Optional[Real] = None,
        sockReadTimeout: Optional[Real] = None,
        sockConnectTimeout: Optional[Real] = None,
    ) -> bool:
        """
        Async test connection to current configured service.

        Args:
            lunaRequestId: External request id. Helps uniquely identifying messages,
                corresponding to particular requests, in system logs.
                Pattern: ^[0-9]{10},[a-f0-9]{8}-[a-f0-9]{4}-4[a-f0-9]{3}-[89ab][a-f0-9]{3}-[a-f0-9]{12}$
            raiseError: if request fails, LunaApiException is raised
            totalTimeout: request processing timeout in seconds.
            connectTimeout: connection timeout seconds.
            sockReadTimeout: socket read timeout
            sockConnectTimeout: socket connect timeout

        Returns:
            *LunaResponse*
        """
        reply = await self.getVersion(
            lunaRequestId=lunaRequestId,
            asyncRequest=True,
            raiseError=raiseError,
            totalTimeout=totalTimeout,
            connectTimeout=connectTimeout,
            sockReadTimeout=sockReadTimeout,
            sockConnectTimeout=sockConnectTimeout,
        )
        return self._connectionChecker(reply)

    def testConnection(
        self,
        lunaRequestId: Optional[str] = None,
        asyncRequest: Optional[bool] = None,
        raiseError: Optional[bool] = False,
        totalTimeout: Optional[Real] = None,
        connectTimeout: Optional[Real] = None,
        sockReadTimeout: Optional[Real] = None,
        sockConnectTimeout: Optional[Real] = None,
        closeSession: bool = True,
    ) -> Union[Awaitable[bool], bool]:
        """
        Test connection to current configured service.

        Args:
            lunaRequestId: External request id. Helps uniquely identifying messages,
                corresponding to particular requests, in system logs.
                Pattern: ^[0-9]{10},[a-f0-9]{8}-[a-f0-9]{4}-4[a-f0-9]{3}-[89ab][a-f0-9]{3}-[a-f0-9]{12}$
            asyncRequest (bool): default mode for request async or blocking
            raiseError: if request fails, LunaApiException is raised
            totalTimeout: request processing timeout in seconds.
            connectTimeout: connection timeout seconds.
            sockReadTimeout: socket read timeout
            sockConnectTimeout: socket connect timeout
            closeSession: whether to close session

        Returns:
            boolean status of establish connection
        """
        coro = self._asyncTestConnection(
            lunaRequestId=lunaRequestId,
            raiseError=raiseError,
            totalTimeout=totalTimeout,
            connectTimeout=connectTimeout,
            sockReadTimeout=sockReadTimeout,
            sockConnectTimeout=sockConnectTimeout,
        )
        if asyncRequest:
            return coro
        loop = asyncio.get_event_loop()
        result = loop.run_until_complete(coro)
        if closeSession:
            loop.run_until_complete(self.close())
        return result

    @property
    def baseUri(self) -> str:
        """
        Property get base part of url.

        >>> self.baseUri
        "http://127.0.0.1:5030/1"
        """
        return f"{self.origin}/{self.api}"

    @property
    def versionUri(self) -> str:
        """
        Property get version uri

        >>> self.versionUri
        "http://127.0.0.1:5030/version"
        """
        return f"{self.origin}/version"

    def getVersion(
        self,
        lunaRequestId: Optional[str] = None,
        asyncRequest: Optional[bool] = None,
        raiseError: Optional[bool] = False,
        totalTimeout: Optional[Real] = None,
        connectTimeout: Optional[Real] = None,
        sockReadTimeout: Optional[Real] = None,
        sockConnectTimeout: Optional[Real] = None,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get version of luna service

        Args:
            lunaRequestId: External request id. Helps uniquely identifying messages, corresponding to particular
                requests, in system logs. Pattern:
                ^[0-9]{10},[a-f0-9]{8}-[a-f0-9]{4}-4[a-f0-9]{3}-[89ab][a-f0-9]{3}-[a-f0-9]{12}$
            asyncRequest: execution in asynchronous mode, disabled by default
            raiseError: if request fails, LunaApiException is raised
            totalTimeout: request processing timeout in seconds.
            connectTimeout: connection timeout seconds.
            sockReadTimeout: socket read timeout
            sockConnectTimeout: socket connect timeout

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with version.
        """
        return self.makeRequest(
            self.versionUri,
            "GET",
            lunaRequestId=lunaRequestId,
            asyncRequest=asyncRequest,
            raiseError=raiseError,
            totalTimeout=totalTimeout,
            connectTimeout=connectTimeout,
            sockReadTimeout=sockReadTimeout,
            sockConnectTimeout=sockConnectTimeout,
        )

    def checkHealth(
        self,
        includeLunaServices: Optional[bool] = None,
        lunaRequestId: Optional[str] = None,
        asyncRequest: Optional[bool] = None,
        raiseError: Optional[bool] = False,
        totalTimeout: Optional[Real] = None,
        connectTimeout: Optional[Real] = None,
        sockReadTimeout: Optional[Real] = None,
        sockConnectTimeout: Optional[Real] = None,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Check health of luna service

        Args:
            includeLunaServices: whether to perform checks for dependencies of the service
            lunaRequestId: External request id. Helps uniquely identifying messages, corresponding to particular
                requests, in system logs. Pattern:
                ^[0-9]{10},[a-f0-9]{8}-[a-f0-9]{4}-4[a-f0-9]{3}-[89ab][a-f0-9]{3}-[a-f0-9]{12}$
            asyncRequest: execution in asynchronous mode, disabled by default
            raiseError: if request fails, LunaApiException is raised
            totalTimeout: request processing timeout in seconds.
            connectTimeout: connection timeout seconds.
            sockReadTimeout: socket read timeout
            sockConnectTimeout: socket connect timeout

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with healthcheck result.
        """
        return self.makeRequest(
            f"{self.origin}/healthcheck",
            "GET",
            queryParams={"include_luna_services": includeLunaServices},
            lunaRequestId=lunaRequestId,
            asyncRequest=asyncRequest,
            raiseError=raiseError,
            totalTimeout=totalTimeout,
            connectTimeout=connectTimeout,
            sockReadTimeout=sockReadTimeout,
            sockConnectTimeout=sockConnectTimeout,
        )

    def getDevelopmentManual(
        self,
        lunaRequestId: Optional[str] = None,
        asyncRequest: Optional[bool] = None,
        raiseError: Optional[bool] = False,
        totalTimeout: Optional[Real] = None,
        connectTimeout: Optional[Real] = None,
        sockReadTimeout: Optional[Real] = None,
        sockConnectTimeout: Optional[Real] = None,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get Development Manual of luna service

        Args:
            lunaRequestId: External request id. Helps uniquely identifying messages, corresponding to particular
                requests, in system logs. Pattern:
                ^[0-9]{10},[a-f0-9]{8}-[a-f0-9]{4}-4[a-f0-9]{3}-[89ab][a-f0-9]{3}-[a-f0-9]{12}$
            asyncRequest: execution in asynchronous mode, disabled by default
            raiseError: if request fails, LunaApiException is raised
            totalTimeout: request processing timeout in seconds.
            connectTimeout: connection timeout seconds.
            sockReadTimeout: socket read timeout
            sockConnectTimeout: socket connect timeout

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return HTML as bytes.
        """
        return makeRequest(
            self.baseUri + "/docs/dev",
            "GET",
            session=self._session,
            headers=self.getRequestIdAsDict(lunaRequestId),
            asyncRequest=self.getAsyncMode(asyncRequest),
            raiseError=raiseError,
            totalTimeout=totalTimeout,
            connectTimeout=connectTimeout,
            sockReadTimeout=sockReadTimeout,
            sockConnectTimeout=sockConnectTimeout,
        )

    def getSpecDocument(
        self,
        acceptHeaders: Optional[dict] = None,
        lunaRequestId: Optional[str] = None,
        asyncRequest: Optional[bool] = None,
        raiseError: Optional[bool] = False,
        totalTimeout: Optional[Real] = None,
        connectTimeout: Optional[Real] = None,
        sockReadTimeout: Optional[Real] = None,
        sockConnectTimeout: Optional[Real] = None,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get OpenApi documentation of luna service

        Args:
            acceptHeaders: Accept request header. By default Accept-Type is text/html.
            lunaRequestId: External request id. Helps uniquely identifying messages, corresponding to particular
                requests, in system logs. Pattern:
                ^[0-9]{10},[a-f0-9]{8}-[a-f0-9]{4}-4[a-f0-9]{3}-[89ab][a-f0-9]{3}-[a-f0-9]{12}$
            asyncRequest: execution in asynchronous mode, disabled by default
            raiseError: if request fails, LunaApiException is raised
            totalTimeout: request processing timeout in seconds.
            connectTimeout: connection timeout seconds.
            sockReadTimeout: socket read timeout
            sockConnectTimeout: socket connect timeout

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return HTML or YAML as bytes.
        """
        standartHeaders = self.getRequestIdAsDict(lunaRequestId)
        standartHeaders.update({"Accept": "text/html"} if acceptHeaders is None else acceptHeaders)

        return makeRequest(
            self.baseUri + "/docs/spec",
            "GET",
            session=self._session,
            headers=standartHeaders,
            asyncRequest=self.getAsyncMode(asyncRequest),
            raiseError=raiseError,
            totalTimeout=totalTimeout,
            connectTimeout=connectTimeout,
            sockReadTimeout=sockReadTimeout,
            sockConnectTimeout=sockConnectTimeout,
        )

    def makeRequest(
        self,
        url: str,
        method: str,
        queryParams: Optional[dict] = None,
        body: Optional[Union[str, bytes, dict, list, aiohttp.Payload, RequestPayload]] = None,
        headers: Optional[dict] = None,
        accept: Optional[str] = None,
        acceptEncoding: Optional[str] = None,
        lunaRequestId: Optional[str] = None,
        raiseError: bool = False,
        asyncRequest: Optional[bool] = None,
        totalTimeout: Optional[Real] = MISSING,
        connectTimeout: Optional[Real] = MISSING,
        sockReadTimeout: Optional[Real] = MISSING,
        sockConnectTimeout: Optional[Real] = MISSING,
        stream: bool = False,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Prepare Luna-Request-Id header and make request.

        Args:
            url: request URL
            method: request method
            queryParams: query parameters
            body: request body
            headers: headers for request
            acceptEncoding: accept-encoding header value
            accept: accept header value
            raiseError: if request fails, LunaApiException is raised
            asyncRequest: execution in asynchronous mode, disabled by default
            lunaRequestId: External request id. Helps uniquely identifying messages, corresponding to particular
                requests, in system logs. Pattern:
                ^[0-9]{10},[a-f0-9]{8}-[a-f0-9]{4}-4[a-f0-9]{3}-[89ab][a-f0-9]{3}-[a-f0-9]{12}$
                        totalTimeout: request processing timeout in seconds.
            totalTimeout: request processing timeout in seconds.
            connectTimeout: connection timeout seconds.
            sockReadTimeout: socket read timeout
            sockConnectTimeout: socket connect timeout
            stream: whether to return request for streaming or consume all of the data in place

        Returns:
            structure with status code, request and decoded Luna API response body is returned.

        """
        requestHeaders = self.getRequestIdAsDict(lunaRequestId)
        if headers:
            requestHeaders.update(headers)

        if totalTimeout is MISSING:
            totalTimeout = self.totalTimeout
        if connectTimeout is MISSING:
            connectTimeout = self.connectTimeout
        if sockReadTimeout is MISSING:
            sockReadTimeout = self.sockReadTimeout
        if sockConnectTimeout is MISSING:
            sockConnectTimeout = self.sockConnectTimeout

        return makeRequest(
            url=url,
            method=method,
            queryParams=queryParams,
            body=body,
            headers=requestHeaders,
            raiseError=raiseError,
            asyncRequest=self.getAsyncMode(asyncRequest),
            session=self._session,
            totalTimeout=totalTimeout,
            connectTimeout=connectTimeout,
            sockReadTimeout=sockReadTimeout,
            sockConnectTimeout=sockConnectTimeout,
            acceptEncoding=acceptEncoding,
            accept=accept,
            stream=stream,
            echo=self.echo,
        )

    def getConfig(
        self,
        acceptHeaders: Optional[dict] = None,
        lunaRequestId: Optional[str] = None,
        asyncRequest: Optional[bool] = None,
        raiseError: Optional[bool] = False,
        totalTimeout: Optional[Real] = None,
        connectTimeout: Optional[Real] = None,
        sockReadTimeout: Optional[Real] = None,
        sockConnectTimeout: Optional[Real] = None,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get config of luna service

        Args:
            acceptHeaders: Accept request header. By default Accept-Type is application/json.
            lunaRequestId: External request id. Helps uniquely identifying messages, corresponding to particular
                requests, in system logs. Pattern:
                ^[0-9]{10},[a-f0-9]{8}-[a-f0-9]{4}-4[a-f0-9]{3}-[89ab][a-f0-9]{3}-[a-f0-9]{12}$
            asyncRequest: execution in asynchronous mode, disabled by default
            raiseError: if request fails, LunaApiException is raised
            totalTimeout: request processing timeout in seconds.
            connectTimeout: connection timeout seconds.
            sockReadTimeout: socket read timeout
            sockConnectTimeout: socket connect timeout

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return HTML or YAML as bytes.
        """
        standardHeaders = self.getRequestIdAsDict(lunaRequestId)
        standardHeaders.update({"Accept": "application/json"} if acceptHeaders is None else acceptHeaders)

        return self.makeRequest(
            self.baseUri + "/config",
            "GET",
            headers=standardHeaders,
            asyncRequest=self.getAsyncMode(asyncRequest),
            raiseError=raiseError,
            totalTimeout=totalTimeout,
            connectTimeout=connectTimeout,
            sockReadTimeout=sockReadTimeout,
            sockConnectTimeout=sockConnectTimeout,
        )
