"""Module contains utils for tasks"""
from typing import Optional

from luna3.common.requests import RequestPayload


def getGCTaskContent(**kwargs) -> dict:
    """
    Get GC task content as dict
    Args:
        **kwargs: kwargs from user request

    Returns:
        dict with content
    """
    target = kwargs.get("target")
    taskContent = {"target": target, "filters": {}}

    handlerId = kwargs.get("handlerId")
    createTimeLt = kwargs.get("createTimeLt")
    insertTimeLt = kwargs.get("insertTimeLt")
    gcAccountId = kwargs.get("gcAccountId")
    createTimeGte = kwargs.get("createTimeGte")
    userData = kwargs.get("userData")
    listId = kwargs.get("listId")
    descriptorVersion = kwargs.get("descriptorVersion")
    descriptorType = kwargs.get("descriptorType")
    storeResults = kwargs.get("storeResults")
    removeSamples = kwargs.get("removeSamples")
    removeImageOrigins = kwargs.get("removeImageOrigins")

    if handlerId is not None:
        taskContent["filters"]["handler_id"] = handlerId
    if createTimeLt is not None:
        taskContent["filters"]["create_time__lt"] = createTimeLt
    if insertTimeLt is not None and target == "events":
        taskContent["filters"]["insert_time__lt"] = insertTimeLt
    if gcAccountId is not None and target in ("events", "faces"):
        taskContent["filters"]["account_id"] = gcAccountId
    if createTimeGte is not None and target == "faces":
        taskContent["filters"]["create_time__gte"] = createTimeGte
    if userData is not None and target == "faces":
        taskContent["filters"]["user_data"] = userData
    if listId is not None and target == "faces":
        taskContent["filters"]["list_id"] = listId
    if descriptorVersion is not None:
        taskContent["filters"]["descriptor_version"] = descriptorVersion
    if descriptorType is not None:
        taskContent["filters"]["descriptor_type"] = descriptorType
    if storeResults is not None:
        taskContent["store_results"] = storeResults
    if removeSamples is not None:
        taskContent["remove_samples"] = removeSamples
    if removeImageOrigins is not None:
        taskContent["remove_image_origins"] = removeImageOrigins

    return taskContent


def getLinkerTaskContent(**kwargs) -> dict:
    """
    Get linker task content as dict
    Args:
        **kwargs: kwargs from user request

    Returns:
        dict with content
    """
    filters = kwargs.get("filters")
    filtersDict = filters.asNoneEmptyDict() if hasattr(filters, "asNoneEmptyDict") else filters

    taskContent = {"objects_type": filtersDict.pop("origin"), "filters": filtersDict}
    listId = kwargs.get("listId")
    createList = kwargs.get("createList")
    listUserData = kwargs.get("listUserData")

    if listId is not None:
        taskContent["list_id"] = listId
    if createList is not None:
        taskContent["create_list"] = createList
    if listUserData is not None:
        taskContent["user_data"] = listUserData

    return taskContent


def getTaskPayload(
    taskType: str, accountId: Optional[str] = None, description: Optional[str] = None, **contentKwargs
) -> RequestPayload:
    """
    Get task data from kwargs
    Args:
        taskType: task type by name (linker/gc)
        accountId: account id
        description: task description
    Keyword Args:
        task content kwargs
    Returns:
        task data
    """
    if taskType == "linker":
        taskContent = getLinkerTaskContent(**contentKwargs)
    elif taskType == "gc":
        taskContent = getGCTaskContent(**contentKwargs)
    else:
        raise RuntimeError("Unsupported task type")

    data = {"content": taskContent}
    if description is not None:
        data["description"] = description

    payload = RequestPayload.buildJson(body=data)
    if accountId is not None:
        payload.applyAccountIdHeader(accountId)

    return payload


def getTaskSchedulePayload(
    taskType: int,
    cron: str,
    cronTz: Optional[str] = None,
    startImmediately: Optional[bool] = None,
    createStopped: Optional[bool] = None,
    accountId: Optional[str] = None,
    **contentKwargs,
) -> RequestPayload:
    """
    Get task schedule payload
    Args:
        taskType: task type
        accountId: account id
        cron: cron string
        cronTz: cron timezone
        startImmediately: start task immediately
        createStopped: create task stopped schedule task
    Keyword Args:
        contentKwargs: task content kwargs
    Returns:
        prepared task schedule payload
    """
    if taskType == 1:
        taskContent = getLinkerTaskContent(**contentKwargs)
    elif taskType == 4:
        taskContent = getGCTaskContent(**contentKwargs)
    else:
        raise RuntimeError("Unsupported task type")

    body = {"task": {"task_type": taskType, "content": taskContent}, "trigger": {"cron": cron}}
    if cronTz is not None:
        body["trigger"]["cron_timezone"] = cronTz
    if startImmediately is not None or createStopped is not None:
        body["behaviour"] = {}
        if startImmediately is not None:
            body["behaviour"]["start_immediately"] = startImmediately
        if createStopped is not None:
            body["behaviour"]["create_stopped"] = createStopped

    payload = RequestPayload.buildJson(body=body)
    if accountId is not None:
        payload.applyAccountIdHeader(accountId)

    return payload
